# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2019
# --------------------------------------------------------------------------

# gendoc: ignore
# This file is generated !
# See script tools/gen_version.py
docplex_version_major = 2
docplex_version_minor = 15
docplex_version_micro = 194
docplex_version_string = '2.15.194'

latest_cplex_major = 12
latest_cplex_minor = 10
