# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

"""
This package contains all modules related to the solving of a CPO model.

It contains in particular the different implementations of the 'solver agents'
allowing to solve on the cloud, locally, etc.
"""