# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

"""
This package contains all modules related to the solving of models expressed
using FlatZinc syntax.

See http://www.minizinc.org/downloads/doc-1.6/flatzinc-spec.pdf for details.
"""