# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

# gendoc: ignore

# import threading
#
# __before_dict_ready = threading.Event()
# __results_ready = threading.Event()
# __before_solve_dict = {}
# __after_solve_dict = {}
# __solve_status = None
# __has_solution = False
# __objective_value = None
# solve_hook = None
